/*
 * Copyright DataStax, Inc.
 *
 * This software can be used solely with DataStax Enterprise. Please consult the license at
 * http://www.datastax.com/terms/datastax-dse-driver-license-terms
 */
package com.datastax.dse.driver.api.core.data.geometry;

import com.datastax.dse.driver.internal.core.data.geometry.DefaultGeometry;
import com.datastax.dse.driver.internal.core.data.geometry.DefaultPoint;
import com.esri.core.geometry.ogc.OGCPoint;
import edu.umd.cs.findbugs.annotations.NonNull;
import java.nio.ByteBuffer;

/**
 * The driver-side representation of DSE's {@code Point}.
 *
 * <p>This is a zero-dimensional object that represents a specific (X,Y) location in a
 * two-dimensional XY-plane. In case of Geographic Coordinate Systems, the X coordinate is the
 * longitude and the Y is the latitude.
 *
 * <p>The default implementation returned by the driver is immutable.
 */
public interface Point extends Geometry {

  /**
   * Creates a point from its <a href="https://en.wikipedia.org/wiki/Well-known_text">Well-known
   * Text</a> (WKT) representation.
   *
   * @param source the Well-known Text representation to parse.
   * @return the point represented by the WKT.
   * @throws IllegalArgumentException if the string does not contain a valid Well-known Text
   *     representation.
   */
  @NonNull
  static Point fromWellKnownText(@NonNull String source) {
    return new DefaultPoint(DefaultGeometry.fromOgcWellKnownText(source, OGCPoint.class));
  }

  /**
   * Creates a point from its <a
   * href="https://en.wikipedia.org/wiki/Well-known_text#Well-known_binary">Well-known Binary</a>
   * (WKB) representation.
   *
   * @param source the Well-known Binary representation to parse.
   * @return the point represented by the WKB.
   * @throws IllegalArgumentException if the provided {@link ByteBuffer} does not contain a valid
   *     Well-known Binary representation.
   */
  @NonNull
  static Point fromWellKnownBinary(@NonNull ByteBuffer source) {
    return new DefaultPoint(DefaultGeometry.fromOgcWellKnownBinary(source, OGCPoint.class));
  }

  /**
   * Creates a point from a <a href="https://tools.ietf.org/html/rfc7946#appendix-A">GeoJSON
   * Point</a> representation.
   *
   * @param source the <a href="https://tools.ietf.org/html/rfc7946#appendix-A">GeoJSON Point</a>
   *     representation to parse.
   * @return the point represented by the <a
   *     href="https://tools.ietf.org/html/rfc7946#appendix-A">GeoJSON Point</a>.
   * @throws IllegalArgumentException if the string does not contain a valid <a
   *     href="https://tools.ietf.org/html/rfc7946#appendix-A">GeoJSON Point</a> representation.
   */
  @NonNull
  static Point fromGeoJson(@NonNull String source) {
    return new DefaultPoint(DefaultGeometry.fromOgcGeoJson(source, OGCPoint.class));
  }

  @NonNull
  static Point fromCoordinates(double x, double y) {
    return new DefaultPoint(x, y);
  }

  /** Returns the X coordinate of this 2D point. */
  double X();

  /** Returns the Y coordinate of this 2D point. */
  double Y();
}
