/*
 * Copyright DataStax, Inc.
 *
 * This software can be used solely with DataStax Enterprise. Please consult the license at
 * http://www.datastax.com/terms/datastax-dse-driver-license-terms
 */
package com.datastax.dse.driver.api.core.metadata.schema;

import com.datastax.oss.driver.api.core.CqlIdentifier;
import com.datastax.oss.driver.api.core.metadata.schema.FunctionSignature;
import com.datastax.oss.driver.api.core.metadata.schema.KeyspaceMetadata;
import com.datastax.oss.driver.api.core.type.DataType;
import com.datastax.oss.driver.shaded.guava.common.collect.ImmutableMap;
import edu.umd.cs.findbugs.annotations.NonNull;
import java.util.Map;
import java.util.Optional;

/**
 * Specialized keyspace metadata for the DSE driver.
 *
 * <p>The only difference with {@link KeyspaceMetadata} is that all sub-elements are specialized for
 * DSE (e.g. {@link #getTables()} returns {@link DseTableMetadata} instances).
 */
public interface DseKeyspaceMetadata extends KeyspaceMetadata {

  @NonNull
  @Override
  Map<CqlIdentifier, ? extends DseTableMetadata> getTables();

  @NonNull
  @Override
  default Optional<? extends DseTableMetadata> getTable(@NonNull CqlIdentifier tableId) {
    return Optional.ofNullable(getTables().get(tableId));
  }

  @NonNull
  @Override
  default Optional<? extends DseTableMetadata> getTable(@NonNull String tableName) {
    return getTable(CqlIdentifier.fromCql(tableName));
  }

  @NonNull
  @Override
  Map<CqlIdentifier, ? extends DseViewMetadata> getViews();

  @NonNull
  @Override
  default Map<CqlIdentifier, ? extends DseViewMetadata> getViewsOnTable(
      @NonNull CqlIdentifier tableId) {
    ImmutableMap.Builder<CqlIdentifier, DseViewMetadata> builder = ImmutableMap.builder();
    for (DseViewMetadata view : getViews().values()) {
      if (view.getBaseTable().equals(tableId)) {
        builder.put(view.getName(), view);
      }
    }
    return builder.build();
  }

  @NonNull
  @Override
  default Optional<? extends DseViewMetadata> getView(@NonNull CqlIdentifier viewId) {
    return Optional.ofNullable(getViews().get(viewId));
  }

  @NonNull
  @Override
  default Optional<? extends DseViewMetadata> getView(@NonNull String viewName) {
    return getView(CqlIdentifier.fromCql(viewName));
  }

  @NonNull
  @Override
  Map<FunctionSignature, ? extends DseFunctionMetadata> getFunctions();

  @NonNull
  @Override
  default Optional<? extends DseFunctionMetadata> getFunction(
      @NonNull FunctionSignature functionSignature) {
    return Optional.ofNullable(getFunctions().get(functionSignature));
  }

  @NonNull
  @Override
  default Optional<? extends DseFunctionMetadata> getFunction(
      @NonNull CqlIdentifier functionId, @NonNull Iterable<DataType> parameterTypes) {
    return Optional.ofNullable(
        getFunctions().get(new FunctionSignature(functionId, parameterTypes)));
  }

  @NonNull
  @Override
  default Optional<? extends DseFunctionMetadata> getFunction(
      @NonNull String functionName, @NonNull Iterable<DataType> parameterTypes) {
    return getFunction(CqlIdentifier.fromCql(functionName), parameterTypes);
  }

  @NonNull
  @Override
  default Optional<? extends DseFunctionMetadata> getFunction(
      @NonNull CqlIdentifier functionId, @NonNull DataType... parameterTypes) {
    return Optional.ofNullable(
        getFunctions().get(new FunctionSignature(functionId, parameterTypes)));
  }

  @NonNull
  @Override
  default Optional<? extends DseFunctionMetadata> getFunction(
      @NonNull String functionName, @NonNull DataType... parameterTypes) {
    return getFunction(CqlIdentifier.fromCql(functionName), parameterTypes);
  }

  @NonNull
  @Override
  Map<FunctionSignature, ? extends DseAggregateMetadata> getAggregates();

  @NonNull
  @Override
  default Optional<? extends DseAggregateMetadata> getAggregate(
      @NonNull FunctionSignature aggregateSignature) {
    return Optional.ofNullable(getAggregates().get(aggregateSignature));
  }

  @NonNull
  @Override
  default Optional<? extends DseAggregateMetadata> getAggregate(
      @NonNull CqlIdentifier aggregateId, @NonNull Iterable<DataType> parameterTypes) {
    return Optional.ofNullable(
        getAggregates().get(new FunctionSignature(aggregateId, parameterTypes)));
  }

  @NonNull
  @Override
  default Optional<? extends DseAggregateMetadata> getAggregate(
      @NonNull String aggregateName, @NonNull Iterable<DataType> parameterTypes) {
    return getAggregate(CqlIdentifier.fromCql(aggregateName), parameterTypes);
  }

  @NonNull
  @Override
  default Optional<? extends DseAggregateMetadata> getAggregate(
      @NonNull CqlIdentifier aggregateId, @NonNull DataType... parameterTypes) {
    return Optional.ofNullable(
        getAggregates().get(new FunctionSignature(aggregateId, parameterTypes)));
  }

  @NonNull
  @Override
  default Optional<? extends DseAggregateMetadata> getAggregate(
      @NonNull String aggregateName, @NonNull DataType... parameterTypes) {
    return getAggregate(CqlIdentifier.fromCql(aggregateName), parameterTypes);
  }
}
