/*
 * Copyright DataStax, Inc.
 *
 * This software can be used solely with DataStax Enterprise. Please consult the license at
 * http://www.datastax.com/terms/datastax-dse-driver-license-terms
 */
package com.datastax.dse.driver.internal.core.metadata;

import com.datastax.dse.driver.api.core.metadata.DseMetadata;
import com.datastax.dse.driver.api.core.metadata.schema.DseKeyspaceMetadata;
import com.datastax.oss.driver.api.core.CqlIdentifier;
import com.datastax.oss.driver.api.core.metadata.Node;
import com.datastax.oss.driver.api.core.metadata.TokenMap;
import com.datastax.oss.driver.api.core.metadata.schema.KeyspaceMetadata;
import com.datastax.oss.driver.internal.core.context.InternalDriverContext;
import com.datastax.oss.driver.internal.core.metadata.DefaultMetadata;
import com.datastax.oss.driver.internal.core.metadata.token.TokenFactory;
import com.datastax.oss.driver.shaded.guava.common.collect.ImmutableMap;
import edu.umd.cs.findbugs.annotations.NonNull;
import java.net.InetSocketAddress;
import java.util.Collections;
import java.util.Map;
import net.jcip.annotations.Immutable;

/**
 * Extends the open-source driver's metadata to return DSE-specific schema metadata.
 *
 * <p>Implementation note: this class performs unchecked casts, relying on the fact that the schema
 * parser builds objects with the appropriate runtime type.
 */
@Immutable
public class DefaultDseMetadata extends DefaultMetadata implements DseMetadata {

  public static DefaultMetadata EMPTY =
      new DefaultDseMetadata(Collections.emptyMap(), Collections.emptyMap(), null);

  protected DefaultDseMetadata(
      Map<InetSocketAddress, Node> nodes,
      Map<CqlIdentifier, ? extends KeyspaceMetadata> keyspaces,
      TokenMap tokenMap) {
    super(nodes, keyspaces, tokenMap);
  }

  @NonNull
  @Override
  @SuppressWarnings("unchecked")
  public Map<CqlIdentifier, ? extends DseKeyspaceMetadata> getKeyspaces() {
    return (Map<CqlIdentifier, ? extends DseKeyspaceMetadata>) keyspaces;
  }

  // Same as the parent method, only specializes the implementation
  @Override
  public DefaultDseMetadata withNodes(
      Map<InetSocketAddress, Node> newNodes,
      boolean tokenMapEnabled,
      boolean tokensChanged,
      TokenFactory tokenFactory,
      InternalDriverContext context) {

    boolean forceFullRebuild = tokensChanged || !newNodes.equals(nodes);

    return new DefaultDseMetadata(
        ImmutableMap.copyOf(newNodes),
        this.keyspaces,
        rebuildTokenMap(
            newNodes, keyspaces, tokenMapEnabled, forceFullRebuild, tokenFactory, context));
  }

  // Same as the parent method, only specializes the implementation
  @Override
  public DefaultDseMetadata withSchema(
      Map<CqlIdentifier, KeyspaceMetadata> newKeyspaces,
      boolean tokenMapEnabled,
      InternalDriverContext context) {
    return new DefaultDseMetadata(
        this.nodes,
        ImmutableMap.copyOf(newKeyspaces),
        rebuildTokenMap(nodes, newKeyspaces, tokenMapEnabled, false, null, context));
  }
}
